@@ -14,10 +14,13 @@ from kivy.properties import ObjectProperty, ListProperty, BooleanProperty, Bound
     NumericProperty
 from kivy.uix.widget import Widget
 from kivy.vector import Vector
+from matplotlib.colors import to_hex
 from matplotlib.backends.backend_agg import FigureCanvasAgg
-from matplotlib.transforms import Bbox
+from matplotlib import cbook
+from weakref import WeakKeyDictionary
 from kivy.metrics import dp
 import numpy as np
+from kivy.utils import get_color_from_hex
 
 class MatplotFigureScatter(Widget):
     """Widget to show a matplotlib figure in kivy.
@@ -46,6 +49,25 @@ class MatplotFigureScatter(Widget):
     pos_y_rect_ver=NumericProperty(0)   
     invert_rect_ver = BooleanProperty(False)
     invert_rect_hor = BooleanProperty(False)
+    legend_instance = ObjectProperty(None, allownone=True)
+    legend_do_scroll_x = BooleanProperty(True)
+    legend_do_scroll_y = BooleanProperty(True) 
+    interactive_axis = BooleanProperty(False)
+    do_pan_x = BooleanProperty(True)
+    do_pan_y = BooleanProperty(True)    
+    do_zoom_x = BooleanProperty(True)
+    do_zoom_y = BooleanProperty(True)
+    fast_draw = BooleanProperty(True) #True will don't draw axis
+    xsorted = BooleanProperty(False) #to manage x sorted data
+    minzoom = NumericProperty(dp(40))
+    multi_xdata = BooleanProperty(False)
+    multi_xdata_res = NumericProperty(dp(20))
+    compare_xdata = BooleanProperty(False)   
+    hover_instance = ObjectProperty(None, allownone=True)
+    nearest_hover_instance = ObjectProperty(None, allownone=True)
+    compare_hover_instance = ObjectProperty(None, allownone=True)
+    disable_mouse_scrolling = BooleanProperty(False) 
+    disable_double_tap = BooleanProperty(False)     
     
     def on_figure(self, obj, value):
         self.figcanvas = _FigureCanvas(self.figure, self)
@@ -61,13 +83,28 @@ class MatplotFigureScatter(Widget):
             ax=self.figure.axes[0]
             patch_cpy=copy.copy(ax.patch)
             patch_cpy.set_visible(False)
-            ax.spines[:].set_zorder(10)
+            for pos in ['right', 'top', 'bottom', 'left']:
+                ax.spines[pos].set_zorder(10)
             patch_cpy.set_zorder(9)
             self.background_patch_copy= ax.add_patch(patch_cpy)
+
+            #set xmin axes attribute
+            self.axes = self.figure.axes[0]
+            
+            #set default xmin/xmax and ymin/ymax
+            self.xmin,self.xmax = self.axes.get_xlim()
+            self.ymin,self.ymax = self.axes.get_ylim()
+
+        if self.legend_instance:
+            self.legend_instance.reset_legend()
+            self.legend_instance=None
             
         # Texture
         self._img_texture = Texture.create(size=(w, h))
 
+        #close last figure in memory (avoid max figure warning)
+        matplotlib.pyplot.close()
+
     def __init__(self, **kwargs):
         super(MatplotFigureScatter, self).__init__(**kwargs)
         
@@ -78,17 +115,14 @@ class MatplotFigureScatter(Widget):
         self.xmax = None
         self.ymin = None
         self.ymax = None
+        self.lines = []
+        self.scatters =[]
         
         #option
-        self.zoompan = None
-        self.fast_draw=True
-        self.draw_left_spline=False #available only when fast_draw is True
         self.touch_mode='pan'
         self.hover_on = False
-        self.xsorted = True #to manage x sorted data (if numpy is used)
-        self.minzoom = dp(40) #minimum pixel distance to apply zoom
-        self.multi_xdata=False
-        self.multi_xdata_res = dp(20)
+        self.cursor_xaxis_formatter=None #used matplotlib formatter to display x cursor value
+        self.cursor_yaxis_formatter=None #used matplotlib formatter to display y cursor value
 
         #zoom box coordonnate
         self.x0_box = None
@@ -107,6 +141,21 @@ class MatplotFigureScatter(Widget):
         #background 
         self.background=None
         self.background_patch_copy=None   
+
+        #manage adjust x and y
+        self.anchor_x = None
+        self.anchor_y = None 
+
+        #manage hover data
+        self.x_hover_data = None
+        self.y_hover_data = None
+
+        #pan management
+        self.first_touch_pan = None
+        
+        #manage back and next event
+        self._nav_stack = cbook.Stack()
+        self.set_history_buttons()   
         
         self.bind(size=self._onSize)
 
@@ -127,11 +176,10 @@ class MatplotFigureScatter(Widget):
         #register lines
         self.lines=lines
                 
-        #white background for blit method (fast draw)
-        props = dict(boxstyle='square',edgecolor='w', facecolor='w', alpha=1.0)
-
         #cursor text
-        self.text = self.axes.text(0.52, 1.01, '', transform=self.axes.transAxes, bbox=props)
+        self.text = self.axes.text(1.0, 1.01, '', 
+                                      transform=self.axes.transAxes,
+                                      ha='right')
 
     def register_scatters(self,scatters:list) -> None:
         """ register scatters method
@@ -188,7 +236,7 @@ class MatplotFigureScatter(Widget):
             if self.multi_xdata:
                 xdata_res, ydata_dump = trans.transform_point((event.x-self.pos[0]+self.multi_xdata_res, 
                                                                event.y - self.pos[1]))
-                delta = xdata_res-xdata 
+                delta = abs(xdata_res-xdata)
                 
             for line in self.lines:
                 #get only visible lines
@@ -328,20 +376,76 @@ class MatplotFigureScatter(Widget):
                         y = self.y_cursor[good_index[idx_best]] 
                     ax=line.axes                     
                     
-                self.set_cross_hair_visible(True)
+                if not self.hover_instance:
+                    self.set_cross_hair_visible(True)
                 
                 # update the cursor x,y data
                 self.horizontal_line.set_ydata(y)
                 self.vertical_line.set_xdata(x)
 
                 #x y label
-                if self.scatter_label  and idx_best > len(good_line)-1: 
-                    if self.multi_xdata:                                
-                        self.text.set_text(f"{self.scatter_label[good_index2_scatter[idx_best]]} x={x}, y={y}")
+                if self.hover_instance:                     
+                    xy_pos = ax.transData.transform([(x,y)]) 
+                    self.x_hover_data = x
+                    self.y_hover_data = y
+                    self.hover_instance.x_hover_pos=float(xy_pos[0][0]) + self.x
+                    self.hover_instance.y_hover_pos=float(xy_pos[0][1]) + self.y
+                    self.hover_instance.show_cursor=True
+                        
+                    if self.cursor_xaxis_formatter:
+                        x = self.cursor_xaxis_formatter.format_data(x)
+                    if self.cursor_yaxis_formatter:
+                        y = self.cursor_yaxis_formatter.format_data(y) 
+                    self.hover_instance.label_x_value=f"{x}"
+                    self.hover_instance.label_y_value=f"{y}"
+            
+                    self.hover_instance.ymin_line = float(ax.bbox.bounds[1])  + self.y
+                    self.hover_instance.ymax_line = float(ax.bbox.bounds[1] + ax.bbox.bounds[3])  + self.y
+
+                    if self.scatter_label  and idx_best > len(good_line)-1: 
+                        if self.multi_xdata:                                
+                            self.hover_instance.custom_label = self.scatter_label[good_index2_scatter[idx_best]]
+                        else:                   
+                            self.hover_instance.custom_label = self.scatter_label[good_index_scatter[idx_best]]
+                    if line:
+                        self.hover_instance.custom_color = get_color_from_hex(to_hex(line.get_color()))
+                    elif scatter:
+                        if self.multi_xdata:
+                            if len(self.x_cursor)==len(scatter.get_facecolors()):
+                                color=scatter.get_facecolors()[good_index2_scatter[idx_best]]
+                            else:
+                                color = scatter.get_facecolors()
+                            self.hover_instance.custom_color = get_color_from_hex(to_hex(color))
+                        else:
+                            if len(self.x_cursor)==len(scatter.get_facecolors()):
+                                color=scatter.get_facecolors()[good_index_scatter[idx_best]]
+                            else:
+                                color = scatter.get_facecolors()                           
+                            self.hover_instance.custom_color = get_color_from_hex(to_hex(color))
+                                                                        
+                    if self.hover_instance.x_hover_pos>self.x+self.axes.bbox.bounds[2] + self.axes.bbox.bounds[0] or \
+                        self.hover_instance.x_hover_pos<self.x+self.axes.bbox.bounds[0] or \
+                        self.hover_instance.y_hover_pos>self.y+self.axes.bbox.bounds[1] + self.axes.bbox.bounds[3] or \
+                        self.hover_instance.y_hover_pos<self.y+self.axes.bbox.bounds[1]:               
+                        self.hover_instance.hover_outside_bound=True
                     else:
-                        self.text.set_text(f"{self.scatter_label[good_index_scatter[idx_best]]} x={x}, y={y}")
-                else:
-                    self.text.set_text(f"x={x}, y={y}")
+                        self.hover_instance.hover_outside_bound=False                      
+                    
+                    return
+                else:                
+
+                    if self.cursor_xaxis_formatter:
+                        x = self.cursor_xaxis_formatter.format_data(x)
+                    if self.cursor_yaxis_formatter:
+                        y = self.cursor_yaxis_formatter.format_data(y) 
+                        
+                    if self.scatter_label  and idx_best > len(good_line)-1: 
+                        if self.multi_xdata:                                
+                            self.text.set_text(f"{self.scatter_label[good_index2_scatter[idx_best]]} x={x}, y={y}")
+                        else:
+                            self.text.set_text(f"{self.scatter_label[good_index_scatter[idx_best]]} x={x}, y={y}")
+                    else:
+                        self.text.set_text(f"x={x}, y={y}")
 
                 #blit method (always use because same visual effect as draw)
                 if self.background is None:
@@ -363,7 +467,13 @@ class MatplotFigureScatter(Widget):
 
             #if touch is too far, hide cross hair cursor
             else:
-                self.set_cross_hair_visible(False)  
+                self.set_cross_hair_visible(False) 
+                if self.hover_instance:
+                    self.hover_instance.x_hover_pos=self.x
+                    self.hover_instance.y_hover_pos=self.y      
+                    self.hover_instance.show_cursor=False
+                    self.x_hover_data = None
+                    self.y_hover_data = None                
 
     def home(self) -> None:
         """ reset data axis
@@ -378,12 +488,89 @@ class MatplotFigureScatter(Widget):
             self.ymax is not None:
                 
             ax = self.axes
-            ax.set_xlim(self.xmin, self.xmax)  
-            ax.set_ylim(self.ymin, self.ymax)                                
+            xleft,xright=ax.get_xlim()
+            ybottom,ytop=ax.get_ylim() 
+            
+            #check inverted data
+            inverted_x = False
+            if xleft>xright:
+                inverted_x=True
+            inverted_y = False
+            if ybottom>ytop:
+                inverted_y=True         
+            
+            if inverted_x:
+                ax.set_xlim(right=self.xmin,left=self.xmax)
+            else:
+                ax.set_xlim(left=self.xmin,right=self.xmax)
+            if inverted_y:
+                ax.set_ylim(top=self.ymin,bottom=self.ymax)
+            else:
+                ax.set_ylim(bottom=self.ymin,top=self.ymax)                                
     
             ax.figure.canvas.draw_idle()
             ax.figure.canvas.flush_events() 
 
+    def back(self, *args):
+        """
+        Move back up the view lim stack.
+        For convenience of being directly connected as a GUI callback, which
+        often get passed additional parameters, this method accepts arbitrary
+        parameters, but does not use them.
+        """
+        self._nav_stack.back()
+        self.set_history_buttons()
+        self._update_view()
+
+    def forward(self, *args):
+        """
+        Move forward in the view lim stack.
+        For convenience of being directly connected as a GUI callback, which
+        often get passed additional parameters, this method accepts arbitrary
+        parameters, but does not use them.
+        """
+        self._nav_stack.forward()
+        self.set_history_buttons()
+        self._update_view()
+
+    def push_current(self):
+       """Push the current view limits and position onto the stack."""
+       self._nav_stack.push(
+           WeakKeyDictionary(
+               {ax: (ax._get_view(),
+                     # Store both the original and modified positions.
+                     (ax.get_position(True).frozen(),
+                      ax.get_position().frozen()))
+                for ax in self.figure.axes}))
+       self.set_history_buttons()       
+
+    def update(self):
+        """Reset the Axes stack."""
+        self._nav_stack.clear()
+        self.set_history_buttons()
+
+    def _update_view(self):
+        """
+        Update the viewlim and position from the view and position stack for
+        each Axes.
+        """
+        nav_info = self._nav_stack()
+        if nav_info is None:
+            return
+        # Retrieve all items at once to avoid any risk of GC deleting an Axes
+        # while in the middle of the loop below.
+        items = list(nav_info.items())
+        for ax, (view, (pos_orig, pos_active)) in items:
+            ax._set_view(view)
+            # Restore both the original and modified positions
+            ax._set_position(pos_orig, 'original')
+            ax._set_position(pos_active, 'active')
+        self.figure.canvas.draw_idle() 
+        self.figure.canvas.flush_events()
+
+    def set_history_buttons(self):
+        """Enable or disable the back/forward button."""
+
     def reset_touch(self) -> None:
         """ reset touch
         
@@ -435,6 +622,26 @@ class MatplotFigureScatter(Widget):
             bytes(self._bitmap), colorfmt="rgba", bufferfmt='ubyte')
         self._img_texture.flip_vertical()
 
+        if self.hover_instance:
+            #update hover pos if needed
+            if self.hover_instance.show_cursor and self.x_hover_data and self.y_hover_data:        
+                xy_pos = self.axes.transData.transform([(self.x_hover_data,self.y_hover_data)]) 
+                self.hover_instance.x_hover_pos=float(xy_pos[0][0]) + self.x
+                self.hover_instance.y_hover_pos=float(xy_pos[0][1]) + self.y
+     
+                # ymin,ymax=self.axes.get_ylim()
+                # ylim_pos = self.axes.transData.transform([(ymin,ymax)])
+                self.hover_instance.ymin_line = float(self.axes.bbox.bounds[1]) + self.y
+                self.hover_instance.ymax_line = float(self.axes.bbox.bounds[1] + self.axes.bbox.bounds[3] )+ self.y
+    
+                if self.hover_instance.x_hover_pos>self.x+self.axes.bbox.bounds[2] + self.axes.bbox.bounds[0] or \
+                    self.hover_instance.x_hover_pos<self.x+self.axes.bbox.bounds[0] or \
+                    self.hover_instance.y_hover_pos>self.y+self.axes.bbox.bounds[1] + self.axes.bbox.bounds[3] or \
+                    self.hover_instance.y_hover_pos<self.y+self.axes.bbox.bounds[1]:               
+                    self.hover_instance.hover_outside_bound=True
+                else:
+                    self.hover_instance.hover_outside_bound=False
+
     def transform_with_touch(self, event):
         """ manage touch behaviour. based on kivy scatter method"""
         # just do a simple one finger drag
@@ -442,9 +649,23 @@ class MatplotFigureScatter(Widget):
 
         if len(self._touches) == self.translation_touches:
             if self.touch_mode=='pan':
+                if self._nav_stack() is None:
+                    self.push_current()                
                 self.apply_pan(self.axes, event)
+
+            if self.touch_mode=='pan_x' or self.touch_mode=='pan_y' \
+                or self.touch_mode=='adjust_x' or self.touch_mode=='adjust_y':
+                if self._nav_stack() is None:
+                    self.push_current()                    
+                self.apply_pan(self.axes, event, mode=self.touch_mode)  
+                
+            elif self.touch_mode=='drag_legend':
+                if self.legend_instance:
+                    self.apply_drag_legend(self.axes, event)                
             
             elif self.touch_mode=='zoombox':
+                if self._nav_stack() is None:
+                    self.push_current()                
                 real_x, real_y = event.x - self.pos[0], event.y - self.pos[1]
                 self.draw_box(event, self.x_init,self.y_init, event.x, real_y)
                 
@@ -497,26 +718,55 @@ class MatplotFigureScatter(Widget):
             changed = True
         return changed
 
+    def on_motion(self,*args):
+        '''Kivy Event to trigger mouse event on motion
+           `enter_notify_event`.
+        '''
+        pos = args[1]
+        newcoord = self.to_widget(pos[0], pos[1])
+        x = newcoord[0]
+        y = newcoord[1]
+        inside = self.collide_point(x,y)
+        if inside: 
+
+            # will receive all motion events.
+            if self.figcanvas and self.hover_instance:
+                #avoid in motion if touch is detected
+                if not len(self._touches)==0:
+                    return
+                FakeEventScatter.x=x
+                FakeEventScatter.y=y
+                self.hover(FakeEventScatter)
+                
     def on_touch_down(self, event):
         """ Manage Mouse/touch press """
         x, y = event.x, event.y
 
-        if self.collide_point(x, y):
+        if self.collide_point(x, y) and self.figure:
+            if self.legend_instance:
+                if self.legend_instance.box.collide_point(x, y):
+                    if self.touch_mode!='drag_legend':
+                        return False   
+                    else:
+                        event.grab(self)
+                        self._touches.append(event)
+                        self._last_touch_pos[event] = event.pos
+                        if len(self._touches)>1:
+                            #new touch, reset background
+                            self.background=None
+                            
+                        return True             
+            
             if event.is_mouse_scrolling:
-                ax = self.axes
-                ax = self.axes
-                self.zoom_factory(event, ax, base_scale=1.2)
+                if not self.disable_mouse_scrolling:
+                    ax = self.axes
+                    ax = self.axes
+                    self.zoom_factory(event, ax, base_scale=1.2)
+                return True
 
-            elif event.is_double_tap:
-                
-                ax = self.axes
-                ax.set_xlim(self.xmin, self.xmax)
-                yoffset = abs(self.ymax - self.ymin) * 0.01
-                ax.set_ylim(self.ymin - yoffset, self.ymax + yoffset)
-                
-                self.reset_touch()
-                ax.figure.canvas.draw_idle()
-                ax.figure.canvas.flush_events() 
+            elif event.is_double_tap:   
+                if not self.disable_double_tap:
+                    self.home()
                 return True
                   
             else:
@@ -547,14 +797,8 @@ class MatplotFigureScatter(Widget):
         x, y = event.x, event.y
 
         if event.is_double_tap:
-            ax = self.axes
-            ax.set_xlim(self.xmin, self.xmax)
-            yoffset = abs(self.ymax - self.ymin) * 0.01
-            ax.set_ylim(self.ymin - yoffset, self.ymax + yoffset)
-
-            self.reset_touch()
-            ax.figure.canvas.draw_idle()
-            ax.figure.canvas.flush_events()                
+            if not self.disable_double_tap:
+                self.home()                
             return True
 
         # scale/translate
@@ -575,7 +819,16 @@ class MatplotFigureScatter(Widget):
             event.ungrab(self)
             del self._last_touch_pos[event]
             self._touches.remove(event)
-
+            if self.touch_mode=='pan' or self.touch_mode=='zoombox' or \
+                self.touch_mode=='pan_x' or self.touch_mode=='pan_y' \
+                or self.touch_mode=='adjust_x' or self.touch_mode=='adjust_y':   
+                self.push_current()
+                if self.interactive_axis:
+                    if self.touch_mode=='pan_x' or self.touch_mode=='pan_y' \
+                        or self.touch_mode=='adjust_x' or self.touch_mode=='adjust_y':
+                        self.touch_mode='pan'                
+                    self.first_touch_pan=None
+                
         x, y = event.x, event.y
         if abs(self._box_size[0]) > 1 or abs(self._box_size[1]) > 1 or self.touch_mode=='zoombox':
             self.reset_box()  
@@ -588,11 +841,14 @@ class MatplotFigureScatter(Widget):
                 return True
             
         # stop propagating if its within our bounds
-        if self.collide_point(x, y):
+        if self.collide_point(x, y) and self.figure:
 
             if self.do_update:
                 self.update_lim()            
 
+            self.anchor_x=None
+            self.anchor_y=None
+            
             ax=self.axes
             self.background=None
             ax.figure.canvas.draw_idle()
@@ -618,8 +874,10 @@ class MatplotFigureScatter(Widget):
         relx = (cur_xlim[1] - xdata) / (cur_xlim[1] - cur_xlim[0])
         rely = (cur_ylim[1] - ydata) / (cur_ylim[1] - cur_ylim[0])
 
-        ax.set_xlim([xdata - new_width * (1 - relx), xdata + new_width * (relx)])
-        ax.set_ylim([ydata - new_height * (1 - rely), ydata + new_height * (rely)])
+        if self.do_zoom_x:
+            ax.set_xlim([xdata - new_width * (1 - relx), xdata + new_width * (relx)])
+        if self.do_zoom_y:
+            ax.set_ylim([ydata - new_height * (1 - rely), ydata + new_height * (rely)])
 
         if self.fast_draw:   
             #use blit method            
@@ -649,22 +907,114 @@ class MatplotFigureScatter(Widget):
             ax.figure.canvas.draw_idle()
             ax.figure.canvas.flush_events()           
 
-    def apply_pan(self, ax, event):
+    def apply_pan(self, ax, event, mode='pan'):
         """ pan method """
         
         trans = ax.transData.inverted()
-        xdata, ydata = trans.transform_point((event.x, event.y))
-        xpress, ypress = trans.transform_point((self._last_touch_pos[event][0], self._last_touch_pos[event][1]))
+        xdata, ydata = trans.transform_point((event.x-self.pos[0], event.y-self.pos[1]))
+        xpress, ypress = trans.transform_point((self._last_touch_pos[event][0]-self.pos[0], self._last_touch_pos[event][1]-self.pos[1]))
         dx = xdata - xpress
         dy = ydata - ypress
 
-        cur_xlim = ax.get_xlim()
-        cur_ylim = ax.get_ylim()
-        cur_xlim -= dx/2
-        cur_ylim -= dy/2
-        ax.set_xlim(cur_xlim)
-        ax.set_ylim(cur_ylim)
+        xleft,xright=self.axes.get_xlim()
+        ybottom,ytop=self.axes.get_ylim()
+        
+        #check inverted data
+        inverted_x = False
+        if xleft>xright:
+            inverted_x=True
+            cur_xlim=(xright,xleft)
+        else:
+            cur_xlim=(xleft,xright)
+        inverted_y = False
+        if ybottom>ytop:
+            inverted_y=True 
+            cur_ylim=(ytop,ybottom)
+        else:
+            cur_ylim=(ybottom,ytop) 
+        
+        if self.interactive_axis and self.touch_mode=='pan' and not self.first_touch_pan=='pan':
+            if (ydata < cur_ylim[0] and not inverted_y) or (ydata > cur_ylim[1] and inverted_y):
+                left_anchor_zone= (cur_xlim[1] - cur_xlim[0])*.2 + cur_xlim[0]
+                right_anchor_zone= (cur_xlim[1] - cur_xlim[0])*.8 + cur_xlim[0]
+                if xdata < left_anchor_zone or xdata > right_anchor_zone:
+                    mode = 'adjust_x'
+                else:
+                    mode = 'pan_x'
+                self.touch_mode = mode
+            elif (xdata < cur_xlim[0] and not inverted_x) or (xdata > cur_xlim[1] and inverted_x):
+                bottom_anchor_zone=  (cur_ylim[1] - cur_ylim[0])*.2 + cur_ylim[0]
+                top_anchor_zone= (cur_ylim[1] - cur_ylim[0])*.8 + cur_ylim[0]               
+                if ydata < bottom_anchor_zone or ydata > top_anchor_zone:
+                    mode = 'adjust_y'
+                else:
+                    mode= 'pan_y' 
+                self.touch_mode = mode
+            else:
+                self.touch_mode = 'pan'        
 
+        if not mode=='pan_y' and not mode=='adjust_y':             
+            if mode=='adjust_x':
+                if self.anchor_x is None:
+                    midpoint= (cur_xlim[1] + cur_xlim[0])/2
+                    if xdata>midpoint:
+                        self.anchor_x='left'
+                    else:
+                        self.anchor_x='right'
+                if self.anchor_x=='left':                
+                    if xdata> cur_xlim[0]:
+                        cur_xlim -= dx/2
+                        if inverted_x:
+                            ax.set_xlim(cur_xlim[1],None)
+                        else:
+                            ax.set_xlim(None,cur_xlim[1])
+                else:
+                    if xdata< cur_xlim[1]:
+                        cur_xlim -= dx/2
+                        if inverted_x:
+                            ax.set_xlim(None,cur_xlim[0])
+                        else:
+                            ax.set_xlim(cur_xlim[0],None)
+            else:
+                cur_xlim -= dx/2
+                if inverted_x:
+                    ax.set_xlim(cur_xlim[1],cur_xlim[0])
+                else:
+                    ax.set_xlim(cur_xlim)
+                
+        if not mode=='pan_x' and not mode=='adjust_x':
+            if mode=='adjust_y':
+                if self.anchor_y is None:
+                    midpoint= (cur_ylim[1] + cur_ylim[0])/2
+                    if ydata>midpoint:
+                        self.anchor_y='top'
+                    else:
+                        self.anchor_y='bottom'               
+                
+                if self.anchor_y=='top':
+                    if ydata> cur_ylim[0]:
+                        cur_ylim -= dy/2 
+                        if inverted_y:
+                            ax.set_ylim(cur_ylim[1],None)
+                        else:
+                            ax.set_ylim(None,cur_ylim[1])
+                else:
+                    if ydata< cur_ylim[1]:
+                        cur_ylim -= dy/2  
+                        if inverted_y:
+                            ax.set_ylim(None,cur_ylim[0]) 
+                        else:
+                            ax.set_ylim(cur_ylim[0],None)
+            else:            
+                cur_ylim -= dy/2
+                if inverted_y:
+                    ax.set_ylim(cur_ylim[1],cur_ylim[0])
+                else:
+                    ax.set_ylim(cur_ylim)
+
+        if self.first_touch_pan is None:
+            self.first_touch_pan=self.touch_mode
+            
         if self.fast_draw: 
             #use blit method
             if self.background is None:
@@ -694,6 +1044,42 @@ class MatplotFigureScatter(Widget):
             ax.figure.canvas.draw_idle()
             ax.figure.canvas.flush_events()
 
+    def apply_drag_legend(self, ax, event):
+        """ drag legend method """
+                        
+        dx = event.x - self._last_touch_pos[event][0]
+        if not self.legend_do_scroll_x:
+            dx=0
+        dy = event.y - self._last_touch_pos[event][1]      
+        if not self.legend_do_scroll_y:
+            dy=0        
+        legend = ax.get_legend()
+        if legend is not None:
+        
+            bbox = legend.get_window_extent()
+            legend_x = bbox.xmin
+            legend_y = bbox.ymin
+               
+            loc_in_canvas = legend_x +dx/2, legend_y+dy/2
+            loc_in_norm_axes = legend.parent.transAxes.inverted().transform_point(loc_in_canvas)
+            legend._loc = tuple(loc_in_norm_axes)
+            
+            #use blit method               
+            if self.background is None:
+                ax.get_legend().set_visible(False)
+                ax.figure.canvas.draw_idle()
+                ax.figure.canvas.flush_events()                   
+                self.background = ax.figure.canvas.copy_from_bbox(ax.figure.bbox)
+                ax.get_legend().set_visible(True)
+            ax.figure.canvas.restore_region(self.background)   
+    
+            ax.draw_artist(legend)
+                
+            ax.figure.canvas.blit(ax.bbox)
+            ax.figure.canvas.flush_events() 
+
+            self.legend_instance.update_size()
+
     def zoom_factory(self, event, ax, base_scale=1.1):
         """ zoom with scrolling mouse method """
 
@@ -724,8 +1110,10 @@ class MatplotFigureScatter(Widget):
         relx = (cur_xlim[1] - xdata) / (cur_xlim[1] - cur_xlim[0])
         rely = (cur_ylim[1] - ydata) / (cur_ylim[1] - cur_ylim[0])
 
-        ax.set_xlim([xdata - new_width * (1 - relx), xdata + new_width * (relx)])
-        ax.set_ylim([ydata - new_height * (1 - rely), ydata + new_height * (rely)])
+        if self.do_zoom_x:
+            ax.set_xlim([xdata - new_width * (1 - relx), xdata + new_width * (relx)])
+        if self.do_zoom_y:
+            ax.set_ylim([ydata - new_height * (1 - rely), ydata + new_height * (rely)])
 
         ax.figure.canvas.draw_idle()
         ax.figure.canvas.flush_events()    
@@ -746,16 +1134,30 @@ class MatplotFigureScatter(Widget):
         hinch = self._height / dpival
         self.figure.set_size_inches(winch, hinch)
         self.figcanvas.resize_event()
-        self.figcanvas.draw()     
-
+        self.figcanvas.draw()
+        if self.legend_instance:
+            self.legend_instance.update_size()        
+        if self.hover_instance:
+            self.hover_instance.figwidth = self.width
+            
     def update_lim(self):
         """ update axis lim if zoombox is used"""
         ax=self.axes
 
         self.do_update=False
         
-        ax.set_xlim(left=min(self.x0_box,self.x1_box),right=max(self.x0_box,self.x1_box))
-        ax.set_ylim(bottom=min(self.y0_box,self.y1_box),top=max(self.y0_box,self.y1_box))
+        #check if inverted axis
+        xleft,xright=self.axes.get_xlim()
+        ybottom,ytop=self.axes.get_ylim()
+        
+        if xright>xleft:
+            ax.set_xlim(left=min(self.x0_box,self.x1_box),right=max(self.x0_box,self.x1_box))
+        else:
+            ax.set_xlim(right=min(self.x0_box,self.x1_box),left=max(self.x0_box,self.x1_box))
+        if ytop>ybottom:
+            ax.set_ylim(bottom=min(self.y0_box,self.y1_box),top=max(self.y0_box,self.y1_box))
+        else:
+            ax.set_ylim(top=min(self.y0_box,self.y1_box),bottom=max(self.y0_box,self.y1_box))
 
     def reset_box(self):
         """ reset zoombox and apply zoombox limit if zoombox option if selected"""
@@ -801,37 +1203,51 @@ class MatplotFigureScatter(Widget):
         trans = self.axes.transData.inverted()
         xdata, ydata = trans.transform_point((event.x-pos_x, event.y-pos_y)) 
 
-        xmin,xmax=self.axes.get_xlim()
-        ymin,ymax=self.axes.get_ylim()
+        xleft,xright=self.axes.get_xlim()
+        ybottom,ytop=self.axes.get_ylim()
         
+        xmax = max(xleft,xright)
+        xmin = min(xleft,xright)
+        ymax = max(ybottom,ytop)
+        ymin = min(ybottom,ytop)
+        
+        #check inverted data
+        inverted_x = False
+        if xleft>xright:
+            inverted_x=True
+        inverted_y = False
+        if ybottom>ytop:
+            inverted_y=True        
+
         x0data, y0data = trans.transform_point((x0-pos_x, y0-pos_y)) 
+         
         if x0data>xmax or x0data<xmin or y0data>ymax or y0data<ymin:
             return
 
         if xdata<xmin:
             x1_min = self.axes.transData.transform([(xmin,ymin)])
-            if x1<x0:
+            if (x1<x0 and not inverted_x) or (x1>x0 and inverted_x):
                 x1=x1_min[0][0]+pos_x
             else:
                 x0=x1_min[0][0]
 
         if xdata>xmax:
             x0_max = self.axes.transData.transform([(xmax,ymin)])
-            if x1>x0:
-                x1=x0_max[0][0]+pos_x
+            if (x1>x0 and not inverted_x) or (x1<x0 and inverted_x):
+                x1=x0_max[0][0]+pos_x 
             else:
                 x0=x0_max[0][0]                  
 
         if ydata<ymin:
             y1_min = self.axes.transData.transform([(xmin,ymin)])
-            if y1<y0:
+            if (y1<y0 and not inverted_y) or (y1>y0 and inverted_y):
                 y1=y1_min[0][1]+pos_y
             else:
                 y0=y1_min[0][1]+pos_y
-                
+
         if ydata>ymax:
             y0_max = self.axes.transData.transform([(xmax,ymax)])
-            if y1>y0:
+            if (y1>y0 and not inverted_y) or (y1<y0 and inverted_y):
                 y1=y0_max[0][1]+pos_y
             else:
                 y0=y0_max[0][1]+pos_y
@@ -846,7 +1262,8 @@ class MatplotFigureScatter(Widget):
             x0_max = self.axes.transData.transform([(xmax,ymin)])
             x1=x0_max[0][0]+pos_x             
 
-            self._alpha_ver=1 
+            self._alpha_ver=1
+            self._alpha_hor=0
              
         elif abs(y1-y0)<dp(20) and abs(x1-x0)>self.minzoom:
             self.pos_x_rect_hor=x0
@@ -859,6 +1276,7 @@ class MatplotFigureScatter(Widget):
             y1=y0_max[0][1]+pos_y         
 
             self._alpha_hor=1
+            self._alpha_ver=0
                         
         else:
             self._alpha_hor=0   
@@ -907,10 +1325,14 @@ class _FigureCanvas(FigureCanvasAgg):
         self.widget.bt_w = w
         self.widget.bt_h = h
         self.widget._draw_bitmap()
+        
+class FakeEventScatter:
+    x:None
+    y:None
 
 from kivy.factory import Factory
 
-Factory.register('MatplotFigure', MatplotFigureScatter)
+Factory.register('MatplotFigureScatter', MatplotFigureScatter)
 
 Builder.load_string('''
 <MatplotFigureScatter>
